/* Copyright (C) 1992 by R.Harmsen
   Interface to FM-synthesizer chip, AdLib or Soundblaster.
   Supports stereo capability of Soundblaster Pro if available.
   (Well, it attempts that, but unsuccessfully)

   Data is from the Dutch boek "Werken met de Sound Blaster",
   which I mentioned here: https://rudhar.com/lingtics/debithet.htm,
   especially around page 214.
   And see http://www.shikadi.net/moddingwiki/OPL_chip, a description
   by Vladimir Arnost, version 1.0, 24 November 1994.

   And see http://bochs.sourceforge.net/techspec/adlib_sb.txt .
   And see http://map.grauw.nl/resources/sound/yamaha_ymf262.pdf .
 */

#include <math.h>
#ifndef GENERATE_TABLE
   #include <dos.h>
   #include "relspeed.h"
#endif
#include "fmchip.h"

#define TEST_REG                0x01
#define COUNT_FAST_REG          0x02
#define COUNT_SLOW_REG          0x03
#define STATUS_REG              0x04

#define MULTI_BASEREG           0x20
#define LEVEL_BASEREG           0x40
#define ATTACK_DECAY_BASEREG    0x60
#define SUSTAIN_RELEASE_BASEREG 0x80
#define FREQ1_BASEREG           0xA0
#define FREQ2_BASEREG           0xB0
#define SYNTH_BASEREG           0xC0
#define WS_BASEREG              0xE0

#define is_percussion() (regimage[0xbd] & 0x20)

fmchip_value_t sbfreq24_value[SBFREQ24_ELEM] =
{
   (0<<10) |  172, (0<<10) |  177, (0<<10) |  183, (0<<10) |  188,
   /* Midinote   2 */
   (0<<10) |  194, (0<<10) |  199, (0<<10) |  205, (0<<10) |  211,
   /* Midinote   4 */
   (0<<10) |  217, (0<<10) |  224, (0<<10) |  230, (0<<10) |  237,
   /* Midinote   6 */
   (0<<10) |  244, (0<<10) |  251, (0<<10) |  258, (0<<10) |  266,
   /* Midinote   8 */
   (0<<10) |  274, (0<<10) |  282, (0<<10) |  290, (0<<10) |  299,
   /* Midinote  10 */
   (0<<10) |  307, (0<<10) |  316, (0<<10) |  326, (0<<10) |  335,
   /* Midinote  12 */
   (0<<10) |  345, (0<<10) |  355, (0<<10) |  365, (0<<10) |  376,
   /* Midinote  14 */
   (0<<10) |  387, (0<<10) |  398, (0<<10) |  410, (0<<10) |  422,
   /* Midinote  16 */
   (0<<10) |  435, (0<<10) |  447, (0<<10) |  460, (0<<10) |  474,
   /* Midinote  18 */
   (0<<10) |  488, (0<<10) |  502, (0<<10) |  517, (0<<10) |  532,
   /* Midinote  20 */
   (0<<10) |  547, (0<<10) |  564, (0<<10) |  580, (0<<10) |  597,
   /* Midinote  22 */
   (0<<10) |  615, (0<<10) |  633, (0<<10) |  651, (0<<10) |  670,
   /* Midinote  24 */
   (0<<10) |  690, (0<<10) |  710, (0<<10) |  731, (0<<10) |  752,
   /* Midinote  26 */
   (0<<10) |  774, (0<<10) |  797, (0<<10) |  820, (0<<10) |  844,
   /* Midinote  28 */
   (0<<10) |  869, (0<<10) |  895, (0<<10) |  921, (0<<10) |  948,
   /* Midinote  30 */
   (0<<10) |  975, (0<<10) | 1004, (1<<10) |  517, (1<<10) |  532,
   /* Midinote  32 */
   (1<<10) |  547, (1<<10) |  564, (1<<10) |  580, (1<<10) |  597,
   /* Midinote  34 */
   (1<<10) |  615, (1<<10) |  633, (1<<10) |  651, (1<<10) |  670,
   /* Midinote  36 */
   (1<<10) |  690, (1<<10) |  710, (1<<10) |  731, (1<<10) |  752,
   /* Midinote  38 */
   (1<<10) |  774, (1<<10) |  797, (1<<10) |  820, (1<<10) |  844,
   /* Midinote  40 */
   (1<<10) |  869, (1<<10) |  895, (1<<10) |  921, (1<<10) |  948,
   /* Midinote  42 */
   (1<<10) |  975, (1<<10) | 1004, (2<<10) |  517, (2<<10) |  532,
   /* Midinote  44 */
   (2<<10) |  547, (2<<10) |  564, (2<<10) |  580, (2<<10) |  597,
   /* Midinote  46 */
   (2<<10) |  615, (2<<10) |  633, (2<<10) |  651, (2<<10) |  670,
   /* Midinote  48 */
   (2<<10) |  690, (2<<10) |  710, (2<<10) |  731, (2<<10) |  752,
   /* Midinote  50 */
   (2<<10) |  774, (2<<10) |  797, (2<<10) |  820, (2<<10) |  844,
   /* Midinote  52 */
   (2<<10) |  869, (2<<10) |  895, (2<<10) |  921, (2<<10) |  948,
   /* Midinote  54 */
   (2<<10) |  975, (2<<10) | 1004, (3<<10) |  517, (3<<10) |  532,
   /* Midinote  56 */
   (3<<10) |  547, (3<<10) |  564, (3<<10) |  580, (3<<10) |  597,
   /* Midinote  58 */
   (3<<10) |  615, (3<<10) |  633, (3<<10) |  651, (3<<10) |  670,
   /* Midinote  60 */
   (3<<10) |  690, (3<<10) |  710, (3<<10) |  731, (3<<10) |  752,
   /* Midinote  62 */
   (3<<10) |  774, (3<<10) |  797, (3<<10) |  820, (3<<10) |  844,
   /* Midinote  64 */
   (3<<10) |  869, (3<<10) |  895, (3<<10) |  921, (3<<10) |  948,
   /* Midinote  66 */
   (3<<10) |  975, (3<<10) | 1004, (4<<10) |  517, (4<<10) |  532,
   /* Midinote  68 */
   (4<<10) |  547, (4<<10) |  564, (4<<10) |  580, (4<<10) |  597,
   /* Midinote  70 */
   (4<<10) |  615, (4<<10) |  633, (4<<10) |  651, (4<<10) |  670,
   /* Midinote  72 */
   (4<<10) |  690, (4<<10) |  710, (4<<10) |  731, (4<<10) |  752,
   /* Midinote  74 */
   (4<<10) |  774, (4<<10) |  797, (4<<10) |  820, (4<<10) |  844,
   /* Midinote  76 */
   (4<<10) |  869, (4<<10) |  895, (4<<10) |  921, (4<<10) |  948,
   /* Midinote  78 */
   (4<<10) |  975, (4<<10) | 1004, (5<<10) |  517, (5<<10) |  532,
   /* Midinote  80 */
   (5<<10) |  547, (5<<10) |  564, (5<<10) |  580, (5<<10) |  597,
   /* Midinote  82 */
   (5<<10) |  615, (5<<10) |  633, (5<<10) |  651, (5<<10) |  670,
   /* Midinote  84 */
   (5<<10) |  690, (5<<10) |  710, (5<<10) |  731, (5<<10) |  752,
   /* Midinote  86 */
   (5<<10) |  774, (5<<10) |  797, (5<<10) |  820, (5<<10) |  844,
   /* Midinote  88 */
   (5<<10) |  869, (5<<10) |  895, (5<<10) |  921, (5<<10) |  948,
   /* Midinote  90 */
   (5<<10) |  975, (5<<10) | 1004, (6<<10) |  517, (6<<10) |  532,
   /* Midinote  92 */
   (6<<10) |  547, (6<<10) |  564, (6<<10) |  580, (6<<10) |  597,
   /* Midinote  94 */
   (6<<10) |  615, (6<<10) |  633, (6<<10) |  651, (6<<10) |  670,
   /* Midinote  96 */
   (6<<10) |  690, (6<<10) |  710, (6<<10) |  731, (6<<10) |  752,
   /* Midinote  98 */
   (6<<10) |  774, (6<<10) |  797, (6<<10) |  820, (6<<10) |  844,
   /* Midinote 100 */
   (6<<10) |  869, (6<<10) |  895, (6<<10) |  921, (6<<10) |  948,
   /* Midinote 102 */
   (6<<10) |  975, (6<<10) | 1004, (7<<10) |  517, (7<<10) |  532,
   /* Midinote 104 */
   (7<<10) |  547, (7<<10) |  564, (7<<10) |  580, (7<<10) |  597,
   /* Midinote 106 */
   (7<<10) |  615, (7<<10) |  633, (7<<10) |  651, (7<<10) |  670,
   /* Midinote 108 */
   (7<<10) |  690, (7<<10) |  710,
};

fmchip_value_t sbfreq53_value[SBFREQ53_ELEM] =
{
   (0<<10) |  172, (0<<10) |  174, (0<<10) |  176, (0<<10) |  179,
   /* Step number    4 ( 4) in 53-EDO */
   (0<<10) |  181, (0<<10) |  183, (0<<10) |  186, (0<<10) |  188,
   /* Step number    8 ( 8) in 53-EDO */
   (0<<10) |  191, (0<<10) |  193, (0<<10) |  196, (0<<10) |  198,
   /* Step number   12 (12) in 53-EDO */
   (0<<10) |  201, (0<<10) |  204, (0<<10) |  206, (0<<10) |  209,
   /* Step number   16 (16) in 53-EDO */
   (0<<10) |  212, (0<<10) |  215, (0<<10) |  217, (0<<10) |  220,
   /* Step number   20 (20) in 53-EDO */
   (0<<10) |  223, (0<<10) |  226, (0<<10) |  229, (0<<10) |  232,
   /* Step number   24 (24) in 53-EDO */
   (0<<10) |  235, (0<<10) |  238, (0<<10) |  241, (0<<10) |  245,
   /* Step number   28 (28) in 53-EDO */
   (0<<10) |  248, (0<<10) |  251, (0<<10) |  254, (0<<10) |  258,
   /* Step number   32 (32) in 53-EDO */
   (0<<10) |  261, (0<<10) |  265, (0<<10) |  268, (0<<10) |  272,
   /* Step number   36 (36) in 53-EDO */
   (0<<10) |  275, (0<<10) |  279, (0<<10) |  283, (0<<10) |  286,
   /* Step number   40 (40) in 53-EDO */
   (0<<10) |  290, (0<<10) |  294, (0<<10) |  298, (0<<10) |  302,
   /* Step number   44 (44) in 53-EDO */
   (0<<10) |  306, (0<<10) |  310, (0<<10) |  314, (0<<10) |  318,
   /* Step number   48 (48) in 53-EDO */
   (0<<10) |  322, (0<<10) |  326, (0<<10) |  331, (0<<10) |  335,
   /* Step number   52 (52) in 53-EDO */
   (0<<10) |  339, (0<<10) |  344, (0<<10) |  348, (0<<10) |  353,
   /* Step number   56 ( 3) in 53-EDO */
   (0<<10) |  358, (0<<10) |  362, (0<<10) |  367, (0<<10) |  372,
   /* Step number   60 ( 7) in 53-EDO */
   (0<<10) |  377, (0<<10) |  382, (0<<10) |  387, (0<<10) |  392,
   /* Step number   64 (11) in 53-EDO */
   (0<<10) |  397, (0<<10) |  402, (0<<10) |  407, (0<<10) |  413,
   /* Step number   68 (15) in 53-EDO */
   (0<<10) |  418, (0<<10) |  424, (0<<10) |  429, (0<<10) |  435,
   /* Step number   72 (19) in 53-EDO */
   (0<<10) |  441, (0<<10) |  447, (0<<10) |  452, (0<<10) |  458,
   /* Step number   76 (23) in 53-EDO */
   (0<<10) |  464, (0<<10) |  471, (0<<10) |  477, (0<<10) |  483,
   /* Step number   80 (27) in 53-EDO */
   (0<<10) |  489, (0<<10) |  496, (0<<10) |  502, (0<<10) |  509,
   /* Step number   84 (31) in 53-EDO */
   (0<<10) |  516, (0<<10) |  522, (0<<10) |  529, (0<<10) |  536,
   /* Step number   88 (35) in 53-EDO */
   (0<<10) |  543, (0<<10) |  550, (0<<10) |  558, (0<<10) |  565,
   /* Step number   92 (39) in 53-EDO */
   (0<<10) |  572, (0<<10) |  580, (0<<10) |  588, (0<<10) |  595,
   /* Step number   96 (43) in 53-EDO */
   (0<<10) |  603, (0<<10) |  611, (0<<10) |  619, (0<<10) |  627,
   /* Step number  100 (47) in 53-EDO */
   (0<<10) |  636, (0<<10) |  644, (0<<10) |  652, (0<<10) |  661,
   /* Step number  104 (51) in 53-EDO */
   (0<<10) |  670, (0<<10) |  679, (0<<10) |  688, (0<<10) |  697,
   /* Step number  108 ( 2) in 53-EDO */
   (0<<10) |  706, (0<<10) |  715, (0<<10) |  724, (0<<10) |  734,
   /* Step number  112 ( 6) in 53-EDO */
   (0<<10) |  744, (0<<10) |  753, (0<<10) |  763, (0<<10) |  773,
   /* Step number  116 (10) in 53-EDO */
   (0<<10) |  784, (0<<10) |  794, (0<<10) |  804, (0<<10) |  815,
   /* Step number  120 (14) in 53-EDO */
   (0<<10) |  826, (0<<10) |  837, (0<<10) |  848, (0<<10) |  859,
   /* Step number  124 (18) in 53-EDO */
   (0<<10) |  870, (0<<10) |  881, (0<<10) |  893, (0<<10) |  905,
   /* Step number  128 (22) in 53-EDO */
   (0<<10) |  917, (0<<10) |  929, (0<<10) |  941, (0<<10) |  953,
   /* Step number  132 (26) in 53-EDO */
   (0<<10) |  966, (0<<10) |  979, (0<<10) |  992, (0<<10) | 1005,
   /* Step number  136 (30) in 53-EDO */
   (0<<10) | 1018, (1<<10) |  516, (1<<10) |  522, (1<<10) |  529,
   /* Step number  140 (34) in 53-EDO */
   (1<<10) |  536, (1<<10) |  543, (1<<10) |  550, (1<<10) |  558,
   /* Step number  144 (38) in 53-EDO */
   (1<<10) |  565, (1<<10) |  572, (1<<10) |  580, (1<<10) |  588,
   /* Step number  148 (42) in 53-EDO */
   (1<<10) |  595, (1<<10) |  603, (1<<10) |  611, (1<<10) |  619,
   /* Step number  152 (46) in 53-EDO */
   (1<<10) |  627, (1<<10) |  636, (1<<10) |  644, (1<<10) |  652,
   /* Step number  156 (50) in 53-EDO */
   (1<<10) |  661, (1<<10) |  670, (1<<10) |  679, (1<<10) |  688,
   /* Step number  160 ( 1) in 53-EDO */
   (1<<10) |  697, (1<<10) |  706, (1<<10) |  715, (1<<10) |  724,
   /* Step number  164 ( 5) in 53-EDO */
   (1<<10) |  734, (1<<10) |  744, (1<<10) |  753, (1<<10) |  763,
   /* Step number  168 ( 9) in 53-EDO */
   (1<<10) |  773, (1<<10) |  784, (1<<10) |  794, (1<<10) |  804,
   /* Step number  172 (13) in 53-EDO */
   (1<<10) |  815, (1<<10) |  826, (1<<10) |  837, (1<<10) |  848,
   /* Step number  176 (17) in 53-EDO */
   (1<<10) |  859, (1<<10) |  870, (1<<10) |  881, (1<<10) |  893,
   /* Step number  180 (21) in 53-EDO */
   (1<<10) |  905, (1<<10) |  917, (1<<10) |  929, (1<<10) |  941,
   /* Step number  184 (25) in 53-EDO */
   (1<<10) |  953, (1<<10) |  966, (1<<10) |  979, (1<<10) |  992,
   /* Step number  188 (29) in 53-EDO */
   (1<<10) | 1005, (1<<10) | 1018, (2<<10) |  516, (2<<10) |  522,
   /* Step number  192 (33) in 53-EDO */
   (2<<10) |  529, (2<<10) |  536, (2<<10) |  543, (2<<10) |  550,
   /* Step number  196 (37) in 53-EDO */
   (2<<10) |  558, (2<<10) |  565, (2<<10) |  572, (2<<10) |  580,
   /* Step number  200 (41) in 53-EDO */
   (2<<10) |  588, (2<<10) |  595, (2<<10) |  603, (2<<10) |  611,
   /* Step number  204 (45) in 53-EDO */
   (2<<10) |  619, (2<<10) |  627, (2<<10) |  636, (2<<10) |  644,
   /* Step number  208 (49) in 53-EDO */
   (2<<10) |  652, (2<<10) |  661, (2<<10) |  670, (2<<10) |  679,
   /* Step number  212 ( 0) in 53-EDO */
   (2<<10) |  688, (2<<10) |  697, (2<<10) |  706, (2<<10) |  715,
   /* Step number  216 ( 4) in 53-EDO */
   (2<<10) |  724, (2<<10) |  734, (2<<10) |  744, (2<<10) |  753,
   /* Step number  220 ( 8) in 53-EDO */
   (2<<10) |  763, (2<<10) |  773, (2<<10) |  784, (2<<10) |  794,
   /* Step number  224 (12) in 53-EDO */
   (2<<10) |  804, (2<<10) |  815, (2<<10) |  826, (2<<10) |  837,
   /* Step number  228 (16) in 53-EDO */
   (2<<10) |  848, (2<<10) |  859, (2<<10) |  870, (2<<10) |  881,
   /* Step number  232 (20) in 53-EDO */
   (2<<10) |  893, (2<<10) |  905, (2<<10) |  917, (2<<10) |  929,
   /* Step number  236 (24) in 53-EDO */
   (2<<10) |  941, (2<<10) |  953, (2<<10) |  966, (2<<10) |  979,
   /* Step number  240 (28) in 53-EDO */
   (2<<10) |  992, (2<<10) | 1005, (2<<10) | 1018, (3<<10) |  516,
   /* Step number  244 (32) in 53-EDO */
   (3<<10) |  522, (3<<10) |  529, (3<<10) |  536, (3<<10) |  543,
   /* Step number  248 (36) in 53-EDO */
   (3<<10) |  550, (3<<10) |  558, (3<<10) |  565, (3<<10) |  572,
   /* Step number  252 (40) in 53-EDO */
   (3<<10) |  580, (3<<10) |  588, (3<<10) |  595, (3<<10) |  603,
   /* Step number  256 (44) in 53-EDO */
   (3<<10) |  611, (3<<10) |  619, (3<<10) |  627, (3<<10) |  636,
   /* Step number  260 (48) in 53-EDO */
   (3<<10) |  644, (3<<10) |  652, (3<<10) |  661, (3<<10) |  670,
   /* Step number  264 (52) in 53-EDO */
   (3<<10) |  679, (3<<10) |  688, (3<<10) |  697, (3<<10) |  706,
   /* Step number  268 ( 3) in 53-EDO */
   (3<<10) |  715, (3<<10) |  724, (3<<10) |  734, (3<<10) |  744,
   /* Step number  272 ( 7) in 53-EDO */
   (3<<10) |  753, (3<<10) |  763, (3<<10) |  773, (3<<10) |  784,
   /* Step number  276 (11) in 53-EDO */
   (3<<10) |  794, (3<<10) |  804, (3<<10) |  815, (3<<10) |  826,
   /* Step number  280 (15) in 53-EDO */
   (3<<10) |  837, (3<<10) |  848, (3<<10) |  859, (3<<10) |  870,
   /* Step number  284 (19) in 53-EDO */
   (3<<10) |  881, (3<<10) |  893, (3<<10) |  905, (3<<10) |  917,
   /* Step number  288 (23) in 53-EDO */
   (3<<10) |  929, (3<<10) |  941, (3<<10) |  953, (3<<10) |  966,
   /* Step number  292 (27) in 53-EDO */
   (3<<10) |  979, (3<<10) |  992, (3<<10) | 1005, (3<<10) | 1018,
   /* Step number  296 (31) in 53-EDO */
   (4<<10) |  516, (4<<10) |  522, (4<<10) |  529, (4<<10) |  536,
   /* Step number  300 (35) in 53-EDO */
   (4<<10) |  543, (4<<10) |  550, (4<<10) |  558, (4<<10) |  565,
   /* Step number  304 (39) in 53-EDO */
   (4<<10) |  572, (4<<10) |  580, (4<<10) |  588, (4<<10) |  595,
   /* Step number  308 (43) in 53-EDO */
   (4<<10) |  603, (4<<10) |  611, (4<<10) |  619, (4<<10) |  627,
   /* Step number  312 (47) in 53-EDO */
   (4<<10) |  636, (4<<10) |  644, (4<<10) |  652, (4<<10) |  661,
   /* Step number  316 (51) in 53-EDO */
   (4<<10) |  670, (4<<10) |  679, (4<<10) |  688, (4<<10) |  697,
   /* Step number  320 ( 2) in 53-EDO */
   (4<<10) |  706, (4<<10) |  715, (4<<10) |  724, (4<<10) |  734,
   /* Step number  324 ( 6) in 53-EDO */
   (4<<10) |  744, (4<<10) |  753, (4<<10) |  763, (4<<10) |  773,
   /* Step number  328 (10) in 53-EDO */
   (4<<10) |  784, (4<<10) |  794, (4<<10) |  804, (4<<10) |  815,
   /* Step number  332 (14) in 53-EDO */
   (4<<10) |  826, (4<<10) |  837, (4<<10) |  848, (4<<10) |  859,
   /* Step number  336 (18) in 53-EDO */
   (4<<10) |  870, (4<<10) |  881, (4<<10) |  893, (4<<10) |  905,
   /* Step number  340 (22) in 53-EDO */
   (4<<10) |  917, (4<<10) |  929, (4<<10) |  941, (4<<10) |  953,
   /* Step number  344 (26) in 53-EDO */
   (4<<10) |  966, (4<<10) |  979, (4<<10) |  992, (4<<10) | 1005,
   /* Step number  348 (30) in 53-EDO */
   (4<<10) | 1018, (5<<10) |  516, (5<<10) |  522, (5<<10) |  529,
   /* Step number  352 (34) in 53-EDO */
   (5<<10) |  536, (5<<10) |  543, (5<<10) |  550, (5<<10) |  558,
   /* Step number  356 (38) in 53-EDO */
   (5<<10) |  565, (5<<10) |  572, (5<<10) |  580, (5<<10) |  588,
   /* Step number  360 (42) in 53-EDO */
   (5<<10) |  595, (5<<10) |  603, (5<<10) |  611, (5<<10) |  619,
   /* Step number  364 (46) in 53-EDO */
   (5<<10) |  627, (5<<10) |  636, (5<<10) |  644, (5<<10) |  652,
   /* Step number  368 (50) in 53-EDO */
   (5<<10) |  661, (5<<10) |  670, (5<<10) |  679, (5<<10) |  688,
   /* Step number  372 ( 1) in 53-EDO */
   (5<<10) |  697, (5<<10) |  706, (5<<10) |  715, (5<<10) |  724,
   /* Step number  376 ( 5) in 53-EDO */
   (5<<10) |  734, (5<<10) |  744, (5<<10) |  753, (5<<10) |  763,
   /* Step number  380 ( 9) in 53-EDO */
   (5<<10) |  773, (5<<10) |  784, (5<<10) |  794, (5<<10) |  804,
   /* Step number  384 (13) in 53-EDO */
   (5<<10) |  815, (5<<10) |  826, (5<<10) |  837, (5<<10) |  848,
   /* Step number  388 (17) in 53-EDO */
   (5<<10) |  859, (5<<10) |  870, (5<<10) |  881, (5<<10) |  893,
   /* Step number  392 (21) in 53-EDO */
   (5<<10) |  905, (5<<10) |  917, (5<<10) |  929, (5<<10) |  941,
   /* Step number  396 (25) in 53-EDO */
   (5<<10) |  953, (5<<10) |  966, (5<<10) |  979, (5<<10) |  992,
   /* Step number  400 (29) in 53-EDO */
   (5<<10) | 1005, (5<<10) | 1018, (6<<10) |  516, (6<<10) |  522,
   /* Step number  404 (33) in 53-EDO */
   (6<<10) |  529, (6<<10) |  536, (6<<10) |  543, (6<<10) |  550,
   /* Step number  408 (37) in 53-EDO */
   (6<<10) |  558, (6<<10) |  565, (6<<10) |  572, (6<<10) |  580,
   /* Step number  412 (41) in 53-EDO */
   (6<<10) |  588, (6<<10) |  595, (6<<10) |  603, (6<<10) |  611,
   /* Step number  416 (45) in 53-EDO */
   (6<<10) |  619, (6<<10) |  627, (6<<10) |  636, (6<<10) |  644,
   /* Step number  420 (49) in 53-EDO */
   (6<<10) |  652, (6<<10) |  661, (6<<10) |  670, (6<<10) |  679,
   /* Step number  424 ( 0) in 53-EDO */
   (6<<10) |  688, (6<<10) |  697, (6<<10) |  706, (6<<10) |  715,
   /* Step number  428 ( 4) in 53-EDO */
   (6<<10) |  724, (6<<10) |  734, (6<<10) |  744, (6<<10) |  753,
   /* Step number  432 ( 8) in 53-EDO */
   (6<<10) |  763, (6<<10) |  773, (6<<10) |  784, (6<<10) |  794,
   /* Step number  436 (12) in 53-EDO */
   (6<<10) |  804, (6<<10) |  815, (6<<10) |  826, (6<<10) |  837,
   /* Step number  440 (16) in 53-EDO */
   (6<<10) |  848, (6<<10) |  859, (6<<10) |  870, (6<<10) |  881,
   /* Step number  444 (20) in 53-EDO */
   (6<<10) |  893, (6<<10) |  905, (6<<10) |  917, (6<<10) |  929,
   /* Step number  448 (24) in 53-EDO */
   (6<<10) |  941, (6<<10) |  953, (6<<10) |  966, (6<<10) |  979,
   /* Step number  452 (28) in 53-EDO */
   (6<<10) |  992, (6<<10) | 1005, (6<<10) | 1018, (7<<10) |  516,
   /* Step number  456 (32) in 53-EDO */
   (7<<10) |  522, (7<<10) |  529, (7<<10) |  536, (7<<10) |  543,
   /* Step number  460 (36) in 53-EDO */
   (7<<10) |  550, (7<<10) |  558, (7<<10) |  565, (7<<10) |  572,
   /* Step number  464 (40) in 53-EDO */
   (7<<10) |  580, (7<<10) |  588, (7<<10) |  595, (7<<10) |  603,
   /* Step number  468 (44) in 53-EDO */
   (7<<10) |  611, (7<<10) |  619, (7<<10) |  627, (7<<10) |  636,
   /* Step number  472 (48) in 53-EDO */
   (7<<10) |  644, (7<<10) |  652, (7<<10) |  661, (7<<10) |  670,
   /* Step number  476 (52) in 53-EDO */
   (7<<10) |  679, (7<<10) |  688, (7<<10) |  697, (7<<10) |  706,
   /* Step number  480 ( 3) in 53-EDO */
   (7<<10) |  715,};

#ifndef GENERATE_TABLE

/*
  Mirror image of the Soundblaster registers, in which the values
  set through this interface are kept. This makes it possible to
  specify "previous value" for the parameters.
 */
static char regimage[256];

struct register_adresses
{
   int regadr;
   int valadr;
};

static struct register_adresses adrs[] =
{
   { 0x388, 0x389 },
   { 0x228, 0x229 },
   { 0x218, 0x219 },
   { 0x228, 0x229 },
   { 0x238, 0x239 },
   #if 0
   /* Don't try this, they may be used for network cards */
   { 0x248, 0x249 },
   { 0x258, 0x259 },
   { 0x268, 0x269 },
   #endif
   { 0    , 0     },
};
static struct register_adresses left_adrs[] =
{
   { 0x240, 0x241 },
   { 0x220, 0x221 },
   /* Don't try this, they may be used for network cards */
   { 0    , 0     },
};
static struct register_adresses rite_adrs[] =
{
   { 0x242, 0x243 },
   { 0x222, 0x223 },
   /* Don't try this, they may be used for network cards */
   { 0    , 0     },
};

/* gcc complains about "not constant, even if I make it a constant */
static struct register_adresses r     = /* adrs[0] */ { 0x388, 0x389 };
static struct register_adresses rleft = /* adrs[0] */ { 0x388, 0x389 };
static struct register_adresses rrite = /* adrs[0] */ { 0x388, 0x389 };
static struct register_adresses rboth = /* adrs[0] */ { 0x388, 0x389 };

/**************************************************************************
   Prototypes of private functions
 *************************************************************************/
static int detect_counter (struct register_adresses *p);
static int chan2reg (int channel, enum modul_carr modcar);
static void write_fmchip (int regis, int value);

/**************************************************************************
   Returns 0 if no chip found, 1 if monophonic AdLib or Soundblaster
   detected, or 2 if stereophonic Soundblaster detected.
 *************************************************************************/
int init_fmchip (void)
{
   struct register_adresses *p, *q;
   int ret_val = 0;

   /* Detect chip */
   for (p = adrs; p->regadr; p++)
   {
      if (detect_counter(p))
      {
         ret_val = 1;
              break;
      }
   }
   /* ... and possible stereo capability.
      For for unknown reasons this causes many computers to hang,
      so assume mono.
    */
   return ret_val;

   for (p = left_adrs; p->regadr; p++)
   {
      if (detect_counter(p))
      {
         ret_val = 2;
         rleft = *p;
         for (q = rite_adrs; q->regadr; q++)
         {
            if (1 || detect_counter(q))
            {
               rrite = *q;
               break;
            }
            else
            {
               /* Left channel found, but right channel is not,
                  must be broke */
               ret_val = 0;
               rleft = rrite = rboth;
            }
         }
      }
      if (ret_val == 2)
         break;
   }
   return ret_val;
}


/**************************************************************************
   See if a Adlib compatible counter can be detected at the given address
   pair. Return 1 if so, else 0.
 *************************************************************************/
static int detect_counter (struct register_adresses *p)
{
   struct register_adresses sav = r;
   int was = 0, is = 0;
   long delay, delay80mu;

   delay80mu = get_relspeed() * 18.2 * 80.0e-6 * 1.9;

   r = *p;

   // Initialize
   write_fmchip(TEST_REG, 0x00);
   /* Mask both counters, and stop them */
   write_fmchip(STATUS_REG, 0x60);
   /* Reset status register */
   write_fmchip(STATUS_REG, 0x80);
   was = inportb(r.regadr);
   /* Set initial value of fast counter to 255,
      so it will overflow at the first tick */
   write_fmchip(COUNT_FAST_REG, 0xff);
   /* Unmask fast counter and start it */
   write_fmchip(STATUS_REG, 0x21);

   /* Delay long enough so the counter ticks at least once.
       It ticks every 80 microseconds */
   for (delay = delay80mu; delay > 0; delay--)
      ;
   is = inportb(r.regadr);

   /* Mask both counters, and stop them */
   write_fmchip(STATUS_REG, 0x60);
   /* Reset status register */
   write_fmchip(STATUS_REG, 0x80);

   r = sav;

   return (was & 0xe0) == 0 && (is & 0xe0) == 0xc0;
}


/**************************************************************************

 *************************************************************************/
void set_envelope (
   int channel,
   enum modul_carr modcar,
   int sustain,
   int attack,
   int decay,
   int sustain_lvl,
   int release)
{
   int reg_offset;
   int multi_reg;
   int attack_decay_reg;
   int sustain_release_reg;
   int value;

   reg_offset = chan2reg(channel, modcar);
   multi_reg = MULTI_BASEREG + reg_offset;
   attack_decay_reg = ATTACK_DECAY_BASEREG + reg_offset;
   sustain_release_reg = SUSTAIN_RELEASE_BASEREG + reg_offset;

   value = regimage[multi_reg];
   if (sustain >= 0)
   {
      if (sustain)
         value |= 0x20;
      else
         value &= ~0x20;
   }

   if (attack < 0)
      attack = regimage[attack_decay_reg] >> 4;
   if (decay < 0)
      decay = regimage[attack_decay_reg] & 0x0f;
   if (sustain < 0)
      sustain_lvl = regimage[sustain_release_reg] >> 4;
   if (release < 0)
      release = regimage[sustain_release_reg] & 0x0f;

   attack      &= 0x0f;
   decay       &= 0x0f;
   sustain_lvl &= 0x0f;
   release     &= 0x0f;

   write_fmchip(multi_reg, value);
   write_fmchip(attack_decay_reg,    (attack      << 4) | decay);
   write_fmchip(sustain_release_reg, (sustain_lvl << 4) | release);
}


/**************************************************************************

 *************************************************************************/
void set_frequency_f (int channel, double freq)
{
   set_frequency_v(channel, freq2value(freq));
}

/**************************************************************************

 *************************************************************************/
void set_frequency_v (int channel, fmchip_value_t value)
{
   int freq1_reg;
   int freq2_reg;

   channel--;
   freq1_reg = FREQ1_BASEREG + channel;
   freq2_reg = FREQ2_BASEREG + channel;

   write_fmchip(freq1_reg, value & 0xff);

   /* Block-value, which set the octave, is in bits 2, 3 and 4.
      Bits 0 and 1 are the upper bits of a ten-bit frequency value.
    */
   write_fmchip(freq2_reg,
      (regimage[freq2_reg] & ~0x1f) | ((value >> 8) & 0x1f));
}
// #ifdef GENERATE_TABLE
#endif


/**************************************************************************
   Conversion between frequency in hertz, and internal soundblaster value,
   consisting of 10 bits, plus a 3 bits octave counter.

   Note 13/14 October 2020: the Dutch book (see top of file) gave the factor
   as 50000, but is it actually approx. 49716, as per
   http://www.shikadi.net/moddingwiki/OPL_chip, which explains deviations
   I noticed with 440 Hz from my tuning fork and a piano.
   This value was historically derived from some ISA bus oscillator signal.
   The difference is no less than almost 10 cents (9.865), or between OPL
   control value 577 (formerly) and 580 (corrected).
 *************************************************************************/
#define SB_FREQ_FACTOR (14318180 / 4 / (double)72)
#include <stdio.h>
fmchip_value_t freq2value (double freq)
{
   int oct;

   freq /= SB_FREQ_FACTOR / (1L << 20);
   oct = log((freq + 0.5) / 512.0) / log(2.0);
   if (oct < 0) oct = 0;
   if (oct > 7) oct = 7;
   freq /= (1 << oct);

   return (int)(freq + 0.5) | (oct << 10);
}

/**************************************************************************
   Conversion between frequency in hertz, and internal soundblaster value,
   consisting of 10 bits, plus a 3 bits octave counter.
 *************************************************************************/
double value2freq (fmchip_value_t value)
{
   int oct = (value & (0x7 << 10)) >> 10;
   int tenbits = value & 0x3ff;

   return SB_FREQ_FACTOR * tenbits / pow(2.0, (20 - oct));
}

#ifndef GENERATE_TABLE

/**************************************************************************

 *************************************************************************/
void set_multiplier (int channel, enum modul_carr modcar, int multiplier)
{
   int reg_offset;
   int regis;
   int value;

   reg_offset = chan2reg(channel, modcar);
   regis = MULTI_BASEREG + reg_offset;

   value = (regimage[regis] & ~0x0f) | (multiplier & 0x0f);

   write_fmchip(regis, value);
}


/**************************************************************************

 *************************************************************************/
void set_level (int channel, enum modul_carr modcar, int level)
{
   int reg_offset;
   int regis;
   int value;

   reg_offset = chan2reg(channel, modcar);
   regis = LEVEL_BASEREG + reg_offset;

   value = (regimage[regis] & ~0x3f) | (level & 0x3f);
   write_fmchip(regis, value);
}


/**************************************************************************

 *************************************************************************/
void set_synth_mode (int channel, enum synth_mode synth)
{
   int regis;
   int value;

   regis = SYNTH_BASEREG + channel - 1;

   value = (regimage[regis] & ~0x01) | (synth == ADDITIVE_SYNTH);
   write_fmchip(regis, value);
}

/**************************************************************************

 *************************************************************************/
void set_fm_feedback (int channel, int feedback)
{
   int regis;
   int value;

   regis = SYNTH_BASEREG + channel - 1;

   value = (regimage[regis] & ~(0x07 << 1)) | ((feedback & 0x07) << 1);
   write_fmchip(regis, value);
}

/**************************************************************************
  Key scale level: ksl may be 0 minimum, and 3 maximum
 *************************************************************************/
void set_ksl (int channel, enum modul_carr modcar, int ksl)
{
   int regis, reg_offset, value;

   reg_offset = chan2reg(channel, modcar);
   regis = LEVEL_BASEREG + reg_offset;

   value = (regimage[regis] & 0x3f) | ((ksl & 0x03) << 6);
   write_fmchip(regis, value);
}


/**************************************************************************
  Key scale rate: ksr may be 0 minimum, and 1 maximum
 *************************************************************************/
void set_ksr (int channel, enum modul_carr modcar, int ksr)
{
   int regis, reg_offset, value;

   reg_offset = chan2reg(channel, modcar);
   regis = MULTI_BASEREG + reg_offset;

   value = (regimage[regis] & 0xef) | ((ksr & 0x01) << 4);
   write_fmchip(regis, value);
}


/**************************************************************************
  Am and vib may be 0 or 1, to indicate amplitude vibrato and/or
  frequency vibrato no/yes.
 *************************************************************************/
void set_am_vib (int channel, enum modul_carr modcar, int am, int vib)
{
   int value, regis, reg_offset;

   reg_offset = chan2reg(channel, modcar);
   regis = MULTI_BASEREG + reg_offset;
   value = regimage[regis];

   if (am != PREV_VAL)
   {
      if (am)
         value |= 0x80;
      else
         value &= ~0x80;
   }
   if (vib != PREV_VAL)
   {
      if (vib)
         value |= 0x40;
      else
         value &= ~0x40;
   }
   write_fmchip(regis, value);
}


/**************************************************************************
  Am and vib may be 0 or 1, to indicate amplitude vibrato and/or
  frequency vibrato intensity low/high.
 *************************************************************************/
void seti_am_vib (int am, int vib)
{
   int value = regimage[0xbd];

   if (am != PREV_VAL)
   {
      if (am)
         value |= 0x80;
      else
         value &= ~0x80;
   }
   if (vib != PREV_VAL)
   {
      if (vib)
         value |= 0x40;
      else
         value &= ~0x40;
   }
   write_fmchip(0xbd, value);
}


/**************************************************************************
  Waveform setting WS: 0 to 3, or 0 to 7 on an OPL3 chip (or emulation)
 *************************************************************************/
void set_ws (int channel, enum modul_carr modcar, int ws)
{
   int regis, reg_offset, value;

   reg_offset = chan2reg(channel, modcar);
   regis = WS_BASEREG + reg_offset;

   /* 13/14 October 2020:
      To enable other waveforms than sines, bit 5 of register 01 must
      set. This information was missing in the Dutch book (see top of
      file), but it is mentioned in
      http://www.shikadi.net/moddingwiki/OPL_chip
      That is why until now it didn't work. */
   if (ws == 0)
   {
      value = regimage[0x1] |  (1 << 5);
      /* Don't reset, because that bit is for all channels, and if
         ws is zero (sine wave) for one channel, it can still be
         nonzero for one or more other channels.z */
      #if 0
      else
         value = regimage[0x1] & ~(1 << 5);
      #endif
      write_fmchip(0x1, value);
   }
   value = (regimage[regis] & ~0x7) | (ws & 0x7);
   write_fmchip(regis, value);
}


/**************************************************************************

 *************************************************************************/
void set_percussion (int yes_no)
{
   int value = regimage[0xbd];

   if (yes_no)
      value |= 0x20;
   else
      value &= ~0x20;

   write_fmchip(0xbd, value);
}


/**************************************************************************

 *************************************************************************/
void note_on (int channel)
{
   int regis;

   note_off(channel);
   regis = FREQ2_BASEREG + channel - 1;
   write_fmchip(regis, regimage[regis] | (1<<5));
}


/**************************************************************************

 *************************************************************************/
void note_off (int channel)
{
   int regis;

   regis = FREQ2_BASEREG + channel - 1;
   write_fmchip(regis, regimage[regis] & ~(1<<5));
}


/**************************************************************************

 *************************************************************************/
void perc_on (int perc_instrument)
{
   int value;

   if (perc_instrument == K_BASS_DRUM)
   {
      int regis;

      regis = FREQ2_BASEREG + 7 - 1;

      write_fmchip(regis, regimage[regis] & ~0x20);
   }
   value = regimage[0xbd];
   write_fmchip(0xbd, value & ~perc_instrument);
   write_fmchip(0xbd, value | perc_instrument);
}


/**************************************************************************
  Set left, right or both channels. If the soundblaster or adlib does
  not have stereo capability, any calls works out as both.
 *************************************************************************/
void set_left_right (enum left_right lrb)
{
   switch (lrb)
   {
      case SB_LEFT:
              r = rleft;
              break;

      case SB_RIGHT:
              r = rrite;
              break;

      case SB_BOTH:
      default:
              r = rboth;
              break;
   }
}

/**************************************************************************
  15 October 2020, 01:30B, I woke up and suddenly understood what this bit
  (bit 6 of register 0x08) is for: when set, the 10 bits value for frequency
  control get an extra 11th bits, implicit and unstored, because it always
  have the value of 1. This improved the resulution, because the useful
  range of 511 to 1023 is replaced by 1024 to 2047.
  Compare the hidden bit (implicit bit) of the significand (mantisse) of
  IEEE 32- and 64-bits floating point formats. Same principle.

  PS, during the day, 15 October: I thought it would make sense if it worked
  that way, but it doesn't. Testing with DosBox 0.74, in OPL2 and OPL3 mode,
  I noticed no effect at all.

  17 October 2020: Now I got it. See http://www.opensound.com/dmguide/dmfm.html
  and other sites with the same text, where it says:

  "kbd_scale is a flag (1-bit) that turns on the keyboard scale rate. If
   kbd_scale is 1 then the attack/decay rates become faster as the pitch
   (fnum + octave) increases."
  "kbd_split This is a flag (1 bit) field which determines the split method to
   select the key scale number. This field is used to select the kbd_scale
   value. Depending on the pitch of the note (octave + fnum), a key scale
   number between 0 and 15 is generated. This key scale number is applied to
   the attack/decay/sustain/release rates depending whether kbd_scale is
   1 or 0. If kbd_split=1 then the key scale number depends on the Most
   Significant Bit (MSB) of the frequency. If kbd_split=0 then the key_scale
   number depends on the the 2nd MSB of the frequency."

  The "key scale number" is the setting KSR.
  So the background is that when bit 6 of register 8 is clear, it is not
  bit 9 of the 10-bit frequency value that determines the faster attach/decay
  for higher notes, but bit 8, because in practice, bit 9 is always 1, for
  optimal frequency accuracy. Values 512-1023 are used, never below.
 *************************************************************************/
void set_nts (int yes_no)
{
   int value;

   if (yes_no)
      value = regimage[0x8] | (1 << 6);
   else
      value = regimage[0x8] & ~(1 << 6);

   write_fmchip(0x08, value);
}


/**************************************************************************
  Return the register number offset belonging to the given channel, and
  whether it is the modulator or the carrier (or either operator when
  in additive mode).
 *************************************************************************/
static int chan2reg (int channel, enum modul_carr modcar)
{
   int regis;

   channel--;

   regis = ((channel / 3) << 3) | (channel % 3);
   if (modcar == CARRIER)
      regis += 3;

   return regis;
}


/**************************************************************************
  Write a value to an FM-chip register.
  The function saves the written value (all 8 bits) in the mirror array.
 *************************************************************************/
#include <stdio.h>
static void write_fmchip (int regis, int value)
{
   long l;
   /* Number of long additions to perform to waste enough time after
      selecting a register until actually writing the value to it.
      Must be at least 3.3 microseconds */
   static long delay1 = -1;
   /* Number of long additions to perform to waste enough time during
      the settling time after writing a value to a port.
      Must be at least 23.3 microseconds. */
   static long delay2;

   if (delay1 < 0)
   {
      delay1 = (long)(get_relspeed() / ((1 / 18.2) / (3.3e-6)) + 1.5);
      delay2 = (long)(delay1 * 23.3 / 3.3 + 1.0);
   }
   regimage[regis] = value;

   outportb(r.regadr, regis);
   /* Settle time at least 3.3 microseconds */
   for (l = 0; l < delay1; l++)
      ;
   outportb(r.valadr, value);
   /* Settle time at least 23.3 microseconds */
   for (l = 0; l < delay2; l++)
      ;
}

#ifdef MAIN
/**************************************************************************
  Testdriver
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

int main (int argc, char **argv)
{
   int channel, base;
   int i, j;
   long l;
   double tones[20];

   init_fmchip();

   randomize();

   if (argc >= 2)
      base = atoi(argv[1]);
   else
      base = rand() % 430 + 50;

   for (j = 0; j < sizeof tones / sizeof tones[0]; j++)
   {
      tones[j] = base * pow(2.0, j / 12.0);
   }

   for (channel = 1; channel <= 9; channel++)
   {
      printf("channel %02d, registers %02x %02x\n", channel,
         chan2reg(channel, MODULATOR),
      chan2reg(channel, CARRIER));
   }

   set_percussion(1);

   (void)getchar();
   printf("base drum\n");
   set_synth_mode(channel, FREQMODU_SYNTH);

   set_envelope(E_BASS_DRUM, MODULATOR, 0, 10, 7, 6, 8);
   set_envelope(E_BASS_DRUM, CARRIER, 0, 10, 7, 6, 8);

   for (i = 0; i < 15; i++)
   {
      set_frequency_f(F_BASS_DRUM, 20 + i * 5);
      perc_on(K_BASS_DRUM);
      (void)getchar();
   }

   printf("tom tom\n");
   set_envelope(E_TOM_TOM, 0, 9, 4, 7, 3);

   for (i = 0; i < 25; i++)
   {
      set_frequency_f(F_BASS_DRUM, 10 + i * 5);
      perc_on(K_BASS_DRUM);
      (void)getchar();
   }

   set_frequency_v(F_TOM_TOM, 60);
   perc_on(K_TOM_TOM);
   (void)getchar();
   for (i = 0; i < 25; i++)
   {
      set_frequency_f(F_TOM_TOM, 10 + i * 5);
      perc_on(K_TOM_TOM);
      (void)getchar();
   }

   printf("snare drum\n");
   set_envelope(E_SNARE_DRUM, 0, 14, 7, 7, 4);
   perc_on(K_SNARE_DRUM);
   (void)getchar();
   perc_on(K_SNARE_DRUM);
   (void)getchar();
   perc_on(K_SNARE_DRUM);
   (void)getchar();
   perc_on(K_SNARE_DRUM);
   (void)getchar();
   perc_on(K_SNARE_DRUM);
   (void)getchar();

   (void)getchar();
   printf("top cymbal\n");
   set_envelope(E_TOP_CYMBAL, 0, 14, 7, 3, 7);
   perc_on(K_TOP_CYMBAL);
   (void)getchar();
   perc_on(K_TOP_CYMBAL);
   (void)getchar();
   perc_on(K_TOP_CYMBAL);
   (void)getchar();
   perc_on(K_TOP_CYMBAL);
   (void)getchar();
   perc_on(K_TOP_CYMBAL);
   (void)getchar();

   printf("hi hat\n");
   set_envelope(E_HI_HAT, 0, 14, 7, 3, 7);
   perc_on(K_HI_HAT);
   (void)getchar();
   perc_on(K_HI_HAT);
   (void)getchar();
   perc_on(K_HI_HAT);
   (void)getchar();
   perc_on(K_HI_HAT);
   (void)getchar();
   perc_on(K_HI_HAT);
   (void)getchar();

   set_percussion(0);

   set_synth_mode(channel, ADDITIVE_SYNTH);

   for (j = 0; j < 24; j++)
   {
      base = tones[rand() % (sizeof tones / sizeof tones[0])];
      for (channel = 1; channel <= 3; channel++)
      {
         set_envelope(channel, MODULATOR, 0, 13, 2, 1 + channel / 3,
            1 + channel / 4 + rand() % 3);
              set_envelope(channel, CARRIER, 0, 13, 2, 1 + channel / 3,
            1 + channel / 4 + rand() % 3);
         set_ksl(channel, MODULATOR, 3);
         set_ksl(channel, CARRIER, 3);
         set_ksr(channel, MODULATOR, 0);
              set_ksr(channel, CARRIER, 0);
         set_level(channel, MODULATOR, 0);
         set_level(channel, CARRIER, 0);
         set_multiplier(channel, MODULATOR, (channel - 1) * 2);
              set_multiplier(channel, CARRIER, (channel - 1) * 2 + 1);
         set_frequency_v(channel, base);

         seti_am_vib (1, 1);
         if (rand() % 737 < 318)
         {
            set_am_vib(channel, MODULATOR, 1, PREV_VAL);
            set_am_vib(channel, CARRIER, 1, PREV_VAL);
            printf("AM on\n");
         }
         else
         {
            set_am_vib(channel, MODULATOR, 0, PREV_VAL);
            set_am_vib(channel, CARRIER, 0, PREV_VAL);
            printf("AM off\n");
         }

         if (rand() % 734 > 317)
         {
            set_am_vib(channel, MODULATOR, PREV_VAL, 1);
            set_am_vib(channel, CARRIER, PREV_VAL, 1);
            printf("VIB on\n");
         }
         else
         {
            set_am_vib(channel, MODULATOR, PREV_VAL, 0);
            set_am_vib(channel, CARRIER, PREV_VAL, 0);
            printf("VIB off\n");
         }
         for (i = 0; i < 4; i++)
         {
            set_ws(channel, MODULATOR, i);
            set_ws(channel, CARRIER, i);
            printf("WS=%d\n", i);
            note_on(channel);
            (void)getchar();
         }
      }
   }
   for (channel = 1; channel <= 9; channel++)
   {
      note_off(channel);
   }

   for (channel = 1; channel <= 9; channel++)
   {
      set_multiplier(channel, MODULATOR, 1);
      set_multiplier(channel, CARRIER  , 1);
      set_level(channel, MODULATOR, 0x00);
      set_level(channel, CARRIER  , 0x00);
   }

   return 0;
}


// #ifdef MAIN
#endif
// #ifndef GENERATE_TABLE
#endif


/**************************************************************************
  Optimized frequency table generator
 *************************************************************************/
#ifdef GENERATE_TABLE
#include <stdio.h>
#include <math.h>

/* Correction 24 September 2020 (yes, some 27 years after first writing
   this source file, and I am still the same programmer, Ruud Harmsen):

   The below value used to be 8.25 Hz, which was 440 * 6/5 / 64. It
   served as the freqency of the lowest C of a nine-octave note table.

   This value of 8.25 can be right as a just temperature frequency, but
   not for equal temperature: it results in 444.03 Hz for the diapason a',
   and all the values in the generated were therefore wrong too.

   So from now on I use a correct value for equidistant tuning. The
   constant formula results in approx. 8.1757989156437 Hz.

   Addition 29 September 2020: In fact this is only correct for 12-
   or 24-step equidistant tuning. If also applied to 53-step equidistant,
   it makes Concert A higher that 440 Hz. So a different calculation for
   low C is needed then.

   Algorithm:
   Midi note #69 = a 440 Hz = index 138 when there are 24 notes per octave.
   69 * 2 / 24 = 5.75 (as is 69 / 12, of course)
   5.75 * 53 = 304.75
   Rounded up, that means index 305 must the 440 Hz ‘a’ note when there are
   53 notes per octave.
   From those values, the frequency of index 0 can be calculated for both
   tables.
 */
#define SUPERLOW_C_24 (440.0 / pow(2.0, 138 / 24.0)) /* 8.1757989156437 Hz */
#define SUPERLOW_C_53 (440.0 / pow(2.0, 305 / 53.0)) /* 8.1491112849877 Hz */

int main (void)
{
   int i, l;

   /* Fill table with values */
   for (i = 0; i < SBFREQ24_ELEM; i++)
   {
      double f = SUPERLOW_C_24 * pow(2.0, i / 24.0);
      sbfreq24_value[i] = freq2value(f);
   }
   /* Write out the table in source code form, to use in this
      same source, see init of sbfreq_value */
   fprintf(stdout, "fmchip_value_t sbfreq24_value[SBFREQ24_ELEM] =\n{\n  ");
   for (l = 1, i = 0; i < SBFREQ24_ELEM; i++)
   {
      fprintf(stdout, " (%1d<<10) | %4d,",
         (int)((sbfreq24_value[i] & (7 << 10)) >> 10),
         (int)( sbfreq24_value[i] & 0x3ff));
      if (++l > 4)
      {
         l = 1;
         fprintf(stdout, "\n   /* Midinote %3d */\n  ", (i+1)>>1);
      }
   }
   fprintf(stdout, "\n};\n\n");

   /* Fill table with values */
   for (i = 0; i < SBFREQ53_ELEM; i++)
   {
      double f = SUPERLOW_C_53 * pow(2.0, i / 53.0);
      sbfreq53_value[i] = freq2value(f);
   }
   fprintf(stdout,
      "fmchip_value_t sbfreq53_value[SBFREQ53_ELEM] =\n{\n  ");
   for (l = 1, i = 0; i < SBFREQ53_ELEM; i++)
   {
      fprintf(stdout, " (%1d<<10) | %4d,",
         (int)((sbfreq53_value[i] & (7 << 10)) >> 10),
         (int)( sbfreq53_value[i] & 0x3ff));
      if (++l > 4)
      {
         l = 1;
         fprintf(stdout, "\n   /* Step number %4d (%2d) in 53-EDO */\n  ",
            i+1, (i+1) % 53);
      }
   }
   fprintf(stdout, "};\n\n");

   return 0;
}

// #ifdef GENERATE_TABLE
#endif
